@@ -3,9 +3,9 @@
  *
  * Common functions for testing TOMOYO Linux's kernel.
  *
- * Copyright (C) 2005-2010  NTT DATA CORPORATION
+ * Copyright (C) 2005-2011  NTT DATA CORPORATION
  *
- * Version: 1.8.0+   2010/12/31
+ * Version: 1.8.2   2011/06/20
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License v2 as published by the
@@ -123,7 +123,7 @@ static pid_t pid = 0;
 
 static void clear_status(void)
 {
-	static const char *keywords[] = {
+	static const char * const keywords[] = {
 		"file::execute",
 		"file::open",
 		"file::create",
@@ -254,12 +254,13 @@ static void ccs_test_init(void)
 	{
 		FILE *fp = fopen("/proc/sys/kernel/osrelease", "r");
 		int version = 0;
-		if (!fp || fscanf(fp, "2.%d.", &version) != 1 || fclose(fp)) {
+		if (!fp || (fscanf(fp, "2.%d.", &version) != 1 &&
+			    fscanf(fp, "%d.", &version) != 1) || fclose(fp)) {
 			fprintf(stderr, "Can't read /proc/sys/kernel/osrelease"
 				"\n");
 			exit(1);
 		}
-		if (version == 6)
+		if (version == 6 || version == 3)
 			is_kernel26 = 1;
 	}
 	{
@@ -298,42 +299,181 @@ static void BUG(const char *fmt, ...)
 		sleep(100);
 }
 
+static char *ccs_freadline(FILE *fp)
+{
+	static char *policy = NULL;
+	int pos = 0;
+	while (1) {
+		static int max_policy_len = 0;
+		const int c = fgetc(fp);
+		if (c == EOF)
+			return NULL;
+		if (pos == max_policy_len) {
+			char *cp;
+			max_policy_len += 4096;
+			cp = realloc(policy, max_policy_len);
+			if (!cp) {
+				BUG("Out of memory");
+				exit(1);
+			}
+			policy = cp;
+		}
+		policy[pos++] = (char) c;
+		if (c == '\n') {
+			policy[--pos] = '\0';
+			break;
+		}
+	}
+	return policy;
+}
+
+static char *ccs_freadline_unpack(FILE *fp)
+{
+	static char *previous_line = NULL;
+	static char *cached_line = NULL;
+	static int pack_start = 0;
+	static int pack_len = 0;
+	if (cached_line)
+		goto unpack;
+	if (!fp)
+		return NULL;
+	{
+		char *pos;
+		unsigned int offset;
+		unsigned int len;
+		char *line = ccs_freadline(fp);
+		if (!line)
+			return NULL;
+		if (sscanf(line, "acl_group %u", &offset) == 1 && offset < 256)
+			pos = strchr(line + 11, ' ');
+		else
+			pos = NULL;
+		if (pos++)
+			offset = pos - line;
+		else
+			offset = 0;
+		if (!strncmp(line + offset, "file ", 5)) {
+			char *cp = line + offset + 5;
+			char *cp2 = strchr(cp + 1, ' ');
+			len = cp2 - cp;
+			if (cp2 && memchr(cp, '/', len)) {
+				pack_start = cp - line;
+				goto prepare;
+			}
+		} else if (!strncmp(line + offset, "network ", 8)) {
+			char *cp = strchr(line + offset + 8, ' ');
+			char *cp2 = NULL;
+			if (cp)
+				cp = strchr(cp + 1, ' ');
+			if (cp)
+				cp2 = strchr(cp + 1, ' ');
+			cp++;
+			len = cp2 - cp;
+			if (cp2 && memchr(cp, '/', len)) {
+				pack_start = cp - line;
+				goto prepare;
+			}
+		}
+		return line;
+prepare:
+		pack_len = len;
+		cached_line = strdup(line);
+		if (!cached_line) {
+			BUG("Out of memory");
+			exit(1);
+		}
+	}
+unpack:
+	{
+		char *line = NULL;
+		char *pos = cached_line + pack_start;
+		char *cp = memchr(pos, '/', pack_len);
+		unsigned int len = cp - pos;
+		free(previous_line);
+		previous_line = NULL;
+		if (!cp) {
+			previous_line = cached_line;
+			cached_line = NULL;
+			line = previous_line;
+		} else if (pack_len == 1) {
+			/* Ignore trailing empty word. */
+			free(cached_line);
+			cached_line = NULL;
+		} else {
+			/* Current string is "abc d/e/f ghi". */
+			line = strdup(cached_line);
+			if (!line) {
+				BUG("Out of memory");
+				exit(1);
+			}
+			previous_line = line;
+			/* Overwrite "abc d/e/f ghi" with "abc d ghi". */
+			memmove(line + pack_start + len, pos + pack_len,
+				strlen(pos + pack_len) + 1);
+			/* Overwrite "abc d/e/f ghi" with "abc e/f ghi". */
+			cp++;
+			memmove(pos, cp, strlen(cp) + 1);
+			/* Forget "d/" component. */
+			pack_len -= len + 1;
+			/* Ignore leading and middle empty word. */
+			if (!len)
+				goto unpack;
+		}
+		return line;
+	}
+}
+
 static int write_domain_policy(const char *policy, int is_delete)
 {
-	FILE *fp = fopen(proc_policy_domain_policy, "r");
-	char buffer[8192];
-	int domain_found = 0;
-	int policy_found = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (!fp) {
-		BUG("Can't read %s", proc_policy_domain_policy);
-		return 0;
+	char *tmp_policy = strdup(policy);
+	if (!tmp_policy) {
+		BUG("Out of memory");
+		exit(1);
 	}
-	if (is_delete)
-		fprintf(domain_fp, "delete ");
-	fprintf(domain_fp, "%s\n", policy);
-	while (fgets(buffer, sizeof(buffer) - 1, fp)) {
-		char *cp = strchr(buffer, '\n');
-		if (cp)
-			*cp = '\0';
-		if (!strncmp(buffer, "<kernel>", 8))
-			domain_found = !strcmp(self_domain, buffer);
-		if (!domain_found)
-			continue;
-		/* printf("<%s>\n", buffer); */
-		if (strcmp(buffer, policy))
-			continue;
-		policy_found = 1;
-		break;
-	}
-	fclose(fp);
-	if (policy_found == is_delete) {
-		if (!strstr(policy, "read/write")) {
+	while (1) {
+		FILE *fp = fopen(proc_policy_domain_policy, "r");
+		_Bool domain_found = 0;
+		_Bool policy_found = 0;
+		if (!fp) {
+			BUG("Can't read %s", proc_policy_domain_policy);
+			return 0;
+		}
+		{
+			char *cp = strrchr(tmp_policy, '\t');
+			if (cp)
+				*cp++ = '\0';
+			else
+				cp = tmp_policy;
+			policy = cp;
+		}
+		if (is_delete)
+			fprintf(domain_fp, "delete ");
+		fprintf(domain_fp, "%s\n", policy);
+		while (1) {
+			char *line = ccs_freadline_unpack(fp);
+			if (!line)
+				break;
+			if (!strncmp(line, "<kernel>", 8))
+				domain_found = !strcmp(self_domain, line);
+			if (!domain_found)
+				continue;
+			/* printf("<%s>\n", buffer); */
+			if (strcmp(line, policy))
+				continue;
+			policy_found = 1;
+			while (ccs_freadline_unpack(NULL));
+			break;
+		}
+		fclose(fp);
+		if (policy_found == is_delete) {
 			BUG("Can't %s %s", is_delete ? "delete" : "append",
 			    policy);
 			return 0;
 		}
+		if (policy == tmp_policy)
+			break;
 	}
+	free(tmp_policy);
 	errno = 0;
 	return 1;
 
@@ -341,32 +481,52 @@ static int write_domain_policy(const char *policy, int is_delete)
 
 static int write_exception_policy(const char *policy, int is_delete)
 {
-	FILE *fp = fopen(proc_policy_exception_policy, "r");
-	char buffer[8192];
-	int policy_found = 0;
-	memset(buffer, 0, sizeof(buffer));
-	if (!fp) {
-		BUG("Can't read %s", proc_policy_exception_policy);
-		return 0;
+	char *tmp_policy = strdup(policy);
+	if (!tmp_policy) {
+		BUG("Out of memory");
+		exit(1);
 	}
-	if (is_delete)
-		fprintf(exception_fp, "delete ");
-	fprintf(exception_fp, "%s\n", policy);
-	while (fgets(buffer, sizeof(buffer) - 1, fp)) {
-		char *cp = strchr(buffer, '\n');
-		if (cp)
-			*cp = '\0';
-		if (strcmp(buffer, policy))
-			continue;
-		policy_found = 1;
-		break;
-	}
-	fclose(fp);
-	if (policy_found == is_delete) {
-		BUG("Can't %s %s", is_delete ? "delete" : "append",
-		    policy);
-		return 0;
+	while (1) {
+		FILE *fp = fopen(proc_policy_exception_policy, "r");
+		_Bool policy_found = 0;
+		if (!fp) {
+			BUG("Can't read %s", proc_policy_exception_policy);
+			return 0;
+		}
+		{
+			char *cp = strrchr(tmp_policy, '\t');
+			if (cp)
+				*cp++ = '\0';
+			else
+				cp = tmp_policy;
+			policy = cp;
+		}
+		if (is_delete)
+			fprintf(exception_fp, "delete ");
+		fprintf(exception_fp, "%s\n", policy);
+		while (1) {
+			char *line = ccs_freadline_unpack(fp);
+			if (!line)
+				break;
+			if (!strncmp(line, "<kernel> ", 9))
+				line += 9;
+			/* printf("<%s>\n", buffer); */
+			if (strcmp(line, policy))
+				continue;
+			policy_found = 1;
+			while (ccs_freadline_unpack(NULL));
+			break;
+		}
+		fclose(fp);
+		if (policy_found == is_delete) {
+			BUG("Can't %s %s", is_delete ? "delete" : "append",
+			    policy);
+			return 0;
+		}
+		if (policy == tmp_policy)
+			break;
 	}
+	free(tmp_policy);
 	errno = 0;
 	return 1;
 
@@ -390,6 +550,8 @@ static int set_profile(const int mode, const char *name)
 		char *cp = strchr(buffer, '\n');
 		if (cp)
 			*cp = '\0';
+		if (!strncmp(buffer, "<kernel> ", 9))
+			memmove(buffer, buffer + 9, strlen(buffer + 9) + 1);
 		if (strncmp(buffer, "255-CONFIG::", 12) ||
 		    strncmp(buffer + 12, name, len) ||
 		    buffer[12 + len] != '=')
